from typing import Any
from .BaseResponseData import BaseResponseData
from dataclasses import dataclass


@dataclass(frozen=True)
class CommentInformTreatment(BaseResponseData):
    should_have_inform_treatment: bool
    text: str
    url: Any  # @TODO
    action_type: Any  # @TODO


@dataclass(frozen=True)
class SharingFrictionInfo(BaseResponseData):
    should_have_sharing_friction: bool
    bloks_app_url: Any  # @TODO
    sharing_friction_payload: Any  # @TODO



from requests import Session, Response
from .containers.Inbox import Inbox
from uuid import uuid4
import string
import random
from .lib.Exceptions import FileTypeError, NetworkError
from json import JSONDecodeError
import time
from pathlib import Path


class Direct:

    session: Session
    device_id: str

    def __init__(self, session: Session, device_id: str) -> None:
        self.session = session
        self.device_id = device_id

    def inbox(self) -> Inbox:
        """
        Fetches all the chats from inbox.
        :return: Inbox object with chat items and some other data
        """

        raise NotImplementedError()

    def send_text(self, text: str, thread_id: int, silently: bool = False) -> bool:
        """
        Sends a text message in a chat
        :param text: Text message
        :param thread_id: Thread ID of chat to send message to
        :param silently: Whether to send message without any notification
        :return: Boolean (Message sent or not)
        """

        mutation_token: str = str().join(random.choices(string.digits, k=19))

        response: Response = self.session.post(
            url="https://i.instagram.com/api/v1/direct_v2/threads/broadcast/text/",
            data={
                "action": "send_item",
                "is_x_transport_forward": False,
                "is_shh_mode": 0,
                "send_silently": silently,
                "thread_ids": f"[{thread_id}]",
                "send_attribution": "inbox",
                "client_context": mutation_token,
                "text": text,
                "device_id": self.device_id,
                "mutation_token": mutation_token,
                "_uuid": str(uuid4()),
                "btt_dual_send": False,
                "is_ae_dual_send": False,
                "offline_threading_id": mutation_token
            }
        )

        try: return response.json().get("status", "") == "ok"

        except JSONDecodeError: return False

    def fb_upload_image(self, media_path: str) -> int:
        """
        Uploads given image to Instagram and returns MediaID.
        :param media_path: Local path of image
        :return: MediaID (Integer)
        """

        media_path: Path = Path(media_path)

        if media_path.suffix not in (".jpg", ".jpeg"):
            raise FileTypeError(
                "Only jpg and jpeg image types can be uploaded."
            )

        with open(media_path, "rb") as file:
            media_data: bytes = file.read()

        upload_name: str = f"{str(int(time.time()) * 1000)}_0_{random.randint(1000000000, 9999999999)}"
        media_length: str = str(len(media_data))

        response: Response = self.session.post(
            url=f"https://rupload.facebook.com/messenger_image/{upload_name}",
            headers={
                "content-length": media_length,
                "content-type": "application/octet-stream",
                "host": "rupload.facebook.com",
                "x-entity-name": upload_name,
                "x-entity-type": "image/jpeg",
                "x-entity-length": media_length,
                "offset": "0",
                "image_type": "FILE_ATTACHMENT"
            },
            data=media_data
        )

        try: return response.json().get("media_id")

        except JSONDecodeError:
            raise NetworkError(
                "Unable to upload image. Make sure there's nothing wrong with the image. "
                "If this problem persists, try using a different account, switch to a different "
                "network, or use reputed proxies."
            )

    def send_photo(self, media_id: int, thread_id: int) -> bool:
        """
        Sends a photo in a chat
        :param media_id: Media ID of photo received by uploading it to 'FB Uploader'
        :param thread_id: Thread ID of chat to send photo to
        :return: Boolean (Photo sent or not)
        """

        mutation_token: str = str().join(random.choices(string.digits, k=19))

        response: Response = self.session.post(
            url="https://i.instagram.com/api/v1/direct_v2/threads/broadcast/photo_attachment/",
            data={
                "action": "send_item",
                "is_x_transport_forward": False,
                "is_shh_mode": 0,
                "thread_ids": f"[{thread_id}]",
                "send_attribution": "inbox",
                "client_context": mutation_token,
                "attachment_fbid": media_id,
                "device_id": self.device_id,
                "mutation_token": mutation_token,
                "_uuid": str(uuid4()),
                "allow_full_aspect_ratio": True,
                "btt_dual_send": False,
                "is_ae_dual_send": False,
                "offline_threading_id": mutation_token
            }
        )

        try: return response.json().get("status", "") == "ok"

        except JSONDecodeError: return False



from dataclasses import dataclass

@dataclass
class BiographyLink:

    """
    Stores a single biography link's information e.g. - Link Title, Link URL, Link ID, etc.
    """

    link_id: int
    url: str
    lynx_url: str
    link_type: str
    title: str
    is_pinned: bool
    open_external_url_with_in_app_browser: bool
    click_id: str



from typing import Any, List
from .Shared import CommentInformTreatment, SharingFrictionInfo
from dataclasses import dataclass
from .BaseResponseData import BaseResponseData


@dataclass(frozen=True)
class PhotoUpload(BaseResponseData):
    raw: dict
    taken_at: int
    pk: str
    id: str
    device_timestamp: int
    client_cache_key: str
    filter_type: int
    caption_is_edited: bool
    like_and_view_counts_disabled: bool
    strong_id__: str
    is_reshare_of_text_post_app_media_in_ig: bool
    has_hidden_comments: bool
    is_post_live_clips_media: bool
    deleted_reason: int
    integrity_review_decision: str
    has_shared_to_fb: int
    is_unified_video: bool
    should_request_ads: bool
    is_visual_reply_commenter_notice_enabled: bool
    commerciality_status: str
    explore_hide_comments: bool
    has_delayed_metadata: bool
    is_quiet_post: bool
    mezql_token: str
    shop_routing_user_id: Any
    can_see_insights_as_brand: bool
    is_organic_product_tagging_eligible: bool
    has_privately_liked: bool
    likers: List
    media_type: int
    code: str
    can_viewer_reshare: bool
    caption: dict  # @TODO caption dataclass
    clips_tab_pinned_user_ids: List  # @TODO
    comment_inform_treatment: CommentInformTreatment
    sharing_friction_info: SharingFrictionInfo
    accessibility_caption: str
    original_media_has_visual_reply_media: bool
    fb_user_tags: Any  # @TODO
    invited_coauthor_producers: List  # @TODO
    can_viewer_save: bool
    is_in_profile_grid: bool
    profile_grid_control_enabled: bool
    featured_products: List  # @TODO
    is_comments_gif_composer_enabled: bool
    highlights_info: Any  # @TODO
    product_suggestions: Any  # @TODO
    user: Any  # @TODO
    image_versions2: Any  # @TODO
    original_width: int
    original_height: int
    enable_media_notes_production: bool
    product_type: str
    is_paid_partnership: bool
    music_metadata: Any  # @TODO
    organic_tracking_token: str
    ig_media_sharing_disabled: bool
    boosted_status: str
    boost_unavailable_identifier: Any  # @TODO
    boost_unavailable_reason: Any  # @TODO
    open_carousel_submission_state: str
    is_open_to_public_submission: bool
    commenting_disabled_for_viewer: bool
    comment_threading_enabled: bool
    max_num_visible_preview_comments: int
    has_more_comments: bool
    preview_comments: List  # @TODO
    comments: List  # @TODO
    comment_count: int
    can_view_more_preview_comments: bool
    hide_view_all_comment_entrypoint: bool
    is_auto_created: bool
    is_cutout_sticker_allowed: bool
    owner: Any  # @TODO



from dataclasses import dataclass
from .BiographyLink import BiographyLink

@dataclass
class Profile:

    """
    Stores someone's profile information e.g. - Full name, Username, Biography, etc.
    """

    full_name: str
    username: str
    is_private: bool
    user_id: str
    is_profile_picture_expansion_enabled: bool
    is_opal_enabled: bool
    is_verified: bool
    profile_pic_url: str
    biography: str
    bio_links: tuple[BiographyLink, ...]
    account_type: int
    is_business: bool
    media_count: int
    following_count: int
    follower_count: int



from ensta.structures.BiographyLink import BiographyLink


def parse_biography_link(information: dict) -> BiographyLink:

    link = BiographyLink(
        link_id=information.get("link_id"),
        url=information.get("url"),
        lynx_url=information.get("lynx_url"),
        link_type=information.get("link_type"),
        title=information.get("title"),
        is_pinned=information.get("is_pinned"),
        open_external_url_with_in_app_browser=information.get("open_external_url_with_in_app_browser"),
        click_id=information.get("click_id")
    )

    return link



from .Exceptions import (
    NetworkError,
    SessionError,
    IdentifierError,
    DevelopmentError,
    AuthenticationError,
    APIError,
    ConversionError,
    FileTypeError,
    RateLimitedError
)



from dataclasses import fields, is_dataclass


class BaseResponseData:

    @classmethod
    def from_data(cls, data):
        parsed_data = {}
        
        for field in fields(cls):
            raw_value = data.get(field.name, None)
            
            if raw_value is None:
                parsed_data[field.name] = None
                
            elif not isinstance(field.type, type):
                parsed_data[field.name] = raw_value
                
            elif issubclass(field.type, BaseResponseData):
                value = field.type.from_data(raw_value)
                parsed_data[field.name] = value
                
            elif is_dataclass(field.type):
                value = field.type(**raw_value)
                parsed_data[field.name] = value
                
            else:
                parsed_data[field.name] = raw_value

        return cls(**parsed_data)

    @classmethod
    def from_response_data(cls, response_data: dict):
        if response_data.get("status", "") != "ok":
            raise Exception(
                "Key 'status' not 'ok' in response JSON. "
                "Please check the images or videos you supplied."
            )
            
        data = response_data.get("media")
        
        if data is None: raise Exception(
            "Either Instagram's Internal Server Error or this library needs to be updated."
        )

        return cls.from_data(data)



import json
import random
import string
import requests
import pyotp
import ntplib
from requests import Session
from requests import Response
from json import JSONDecodeError
from .PasswordEncryption import PasswordEncryption
from .lib.Exceptions import (AuthenticationError, NetworkError)
from .SessionHost import SessionHost


def new_session_id(
    user_identifier: str,  # Username or Email
    password: str,
    proxy: dict[str, str],
    totp_token: str = None
) -> str:
    
    request_session: Session = requests.Session()
    request_session.headers["user-agent"] = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 " \
                                            "(KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36"

    if proxy is not None: request_session.proxies.update(proxy)

    encryption = PasswordEncryption(request_session)
    encrypted_password = encryption.encrypt(password)

    data: dict = {
        "enc_password": encrypted_password,
        "optIntoOneTap": False,
        "queryParams": "{}",
        "trustedDeviceRecords": "{}",
        "username": user_identifier  # Accepts email as well
    }

    csrf_token: str = "".join(random.choices(string.ascii_letters + string.digits, k=32))

    headers: dict = {
        "accept": "*/*",
        "accept-language": "en-US,en;q=0.9",
        "content-type": "application/x-www-form-urlencoded",
        "dpr": "1.30208",
        "sec-ch-prefers-color-scheme": "dark",
        "sec-ch-ua": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
                     "Chrome/119.0.0.0 Safari/537.36",
        "sec-ch-ua-full-version-list": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                                       "(KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-model": "\"\"",
        "sec-ch-ua-platform": "\"Windows\"",
        "sec-ch-ua-platform-version": "\"15.0.0\"",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "viewport-width": "1475",
        "x-asbd-id": "129477",
        "x-csrftoken": csrf_token,
        "x-ig-app-id": "936619743392459",
        "x-ig-www-claim": "0",
        "x-instagram-ajax": "1009977574",
        "x-requested-with": "XMLHttpRequest",
        "x-web-device-id": "25532C62-8BBC-4927-B6C5-02631D6E05BF",
        "cookie": f"dpr=1.3020833730697632; csrftoken={csrf_token}",
        "Referer": "https://www.instagram.com/",
        "Referrer-Policy": "strict-origin-when-cross-origin"
    }

    http_response: Response = request_session.post(
        url="https://www.instagram.com/api/v1/web/accounts/login/ajax/",
        data=data,
        headers=headers
    )

    try:
        response_json: dict = http_response.json()

        if response_json.get("status", "") != "ok":

            verification_code: int | None = None

            if response_json.get("two_factor_required", False) is True:
            
                if response_json.get("two_factor_info", {}).get("totp_two_factor_on", False) is False:

                    if response_json.get("two_factor_info", {}).get("sms_two_factor_on", False) is False:

                        raise AuthenticationError(
                            "Some other 2FA method is enabled. Only TOTP-based"
                            " (Authenticator App) and SMS-based two factor is supported."
                        )

                    verification_code: int = int(input("SMS 2FA enabled. Enter OTP: "))
    
                if totp_token is None and verification_code is None:
                    raise AuthenticationError("Two-factor is enabled. Please provide the totp_token while logging in.")

                else:
                    tf_data: dict = {
                        "queryParams": '{"next":"/"}',
                        "trust_signal": True,
                        "identifier": response_json.get("two_factor_info", {}).get("two_factor_identifier"),
                        "verification_method": "1",
                        "username": user_identifier,
                        "verificationCode": verification_code if verification_code is not None else

                        pyotp.TOTP(totp_token).at(
                            int(
                                ntplib.NTPClient().request(
                                    "time.google.com",
                                    version=3
                                ).tx_time
                            )
                        )
                    }

                    tf_response: Response = request_session.post(
                        url="https://www.instagram.com/api/v1/web/accounts/login/ajax/two_factor/",
                        data=tf_data,
                        headers={
                            "sec-ch-prefers-color-scheme": "dark",
                            "sec-ch-ua": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                                         "AppleWebKit/537.36 (KHTML, like Gecko) "
                                         "Chrome/119.0.0.0 Safari/537.36",
                            "sec-ch-ua-full-version-list": "Mozilla/5.0 (Windows NT 10.0; Win64;"
                                                           " x64) AppleWebKit/537.36 "
                                                           "(KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
                            "sec-ch-ua-mobile": "?0",
                            "sec-ch-ua-model": "\"\"",
                            "sec-ch-ua-platform": "\"Windows\"",
                            "sec-ch-ua-platform-version": "\"15.0.0\"",
                            "sec-fetch-dest": "empty",
                            "sec-fetch-mode": "cors",
                            "sec-fetch-site": "same-origin",
                            'Accept': '*/*',
                            'Accept-Language': 'en-US,en;q=0.5',
                            'X-Mid': response_json.get("two_factor_info", {}).get("device_id"),
                            'X-CSRFToken': csrf_token,
                            "x-instagram-ajax": "1009977574",
                            "x-ig-app-id": "936619743392459",
                            'X-ASBD-ID': '129477',
                            'X-IG-WWW-Claim': '0',
                            "x-web-device-id": "25532C62-8BBC-4927-B6C5-02631D6E05BF",
                            'Content-Type': 'application/x-www-form-urlencoded',
                            'X-Requested-With': 'XMLHttpRequest',
                            'Origin': 'https://www.instagram.com',
                            'DNT': '1',
                            'Sec-GPC': '1',
                            'Connection': 'keep-alive',
                            'Referer': 'https://www.instagram.com/accounts/login/two_factor?next=%2F',
                            'Sec-Fetch-Dest': 'empty',
                            'Sec-Fetch-Mode': 'cors',
                            'Sec-Fetch-Site': 'same-origin',
                        },
                        cookies={
                            "ig_did": response_json.get("ig_did", ""),
                            "mid": response_json.get("two_factor_info", {}).get("device_id"),
                            "csrftoken": csrf_token
                        }
                    )
                     
                    if "Oops, an error occurred." in tf_response.text:

                        raise AuthenticationError(
                            "IP temporarily banned most probably due to too many login requests."
                            " Please try again later or use proxies."
                        )
                    
                    try:
                        tf_response_json: dict = tf_response.json()
                    
                        if tf_response_json.get("status", "") != "ok" \
                                or tf_response_json.get("authenticated", False) is False:

                            raise AuthenticationError(
                                "Couldn't log in through 2FA. Most probably your totp_token is incorrect."
                            )
                        
                        session_id: str = tf_response.cookies.get("sessionid", "")
                        rur: str = tf_response.cookies.get("rur", "")
                        mid: str = response_json.get("two_factor_info", {}).get("device_id")
                        user_id: str = tf_response_json.get("userId", "")
                        ig_did: str = tf_response.cookies.get("ig_did", "")
    
                        if session_id == "" or user_id == "":

                            raise AuthenticationError(
                                "2FA authentication response didn't return a valid session_id or user_id."
                            )

                        return json.dumps({
                            "session_id": session_id,
                            "rur": rur,
                            "mid": mid,
                            "user_id": user_id,
                            "ig_did": ig_did,
                            "identifier": user_identifier,
                            "username": SessionHost(
                                json.dumps({
                                    "session_id": session_id,
                                    "rur": rur,
                                    "mid": mid,
                                    "user_id": user_id,
                                    "ig_did": ig_did,
                                })
                            ).private_info().username
                        })

                    except JSONDecodeError:
                        raise NetworkError(
                            "Response got while logging in was not a valid "
                            "json. Are you able to visit Instagram on the web?"
                        )
            
            raise AuthenticationError(
                "Either user doesn't exist or your password is too weak (change it to a stronger one)."
            )

        if response_json.get("user", False) is False or response_json.get("authenticated", False) is False:
            raise AuthenticationError("Invalid password.")

        session_id: str = http_response.cookies.get("sessionid", "")
        rur: str = http_response.cookies.get("rur", "")
        mid: str = http_response.cookies.get("mid", "")
        user_id: str = response_json.get("userId", "")
        ig_did: str = response_json.get("ig_did", "")

        if session_id == "" or user_id == "": raise AuthenticationError("Unable to login.")

        return json.dumps({
            "session_id": session_id,
            "rur": rur,
            "mid": mid,
            "user_id": user_id,
            "ig_did": ig_did,
            "identifier": user_identifier,
            "username": SessionHost(
                json.dumps({
                    "session_id": session_id,
                    "rur": rur,
                    "mid": mid,
                    "user_id": user_id,
                    "ig_did": ig_did,
                })
            ).private_info().username
        })

    except JSONDecodeError:
        raise NetworkError("Response got while logging in was not a valid "
                           "json. Are you able to visit Instagram on the web?")



import os
import json
from .SessionHost import SessionHost
from .lib.Exceptions import SessionError
from .Authentication import new_session_id


# noinspection PyMissingConstructor
class Host(SessionHost):

    DEFAULT_FILE: str = "ensta-web-session.txt"

    identifier: str = None
    password: str = None

    save: any = None
    load: any = None
    file: str | None = None

    proxy: dict[str, str] = None
    totp_token: str = None

    def __init__(
        self,
        identifier: str,
        password: str,
        file: str = None,
        save: any = None,
        load: any = None,
        proxy: dict[str, str] = None,
        totp_token: str = None
    ) -> None:

        """
        Login using your username/email and password.
        :param identifier: Your Instagram Username or Email
        :param password: Your Instagram Password
        :param file: (Optional) Name of file to store session_data in. Eg: "session.txt"
        :param save: (Optional) Name of a function which will be called to save the generated session_data
        :param load: (Optional) Name of a function which will be called to receive the session_data
        :param proxy: (Optional) JSON Object of proxy to be used. See https://github.com/diezo/ensta
        :param totp_token: (Optional) Your TOTP Key generated by Instagram while setting up 2FA (If 2FA is turned on)
        """

        self.identifier: str = identifier
        self.password: str = password
        self.file: str = file
        self.save: any = save
        self.load: any = load
        self.proxy: dict[str, str] = proxy
        self.totp_token = totp_token

        if self.file is None and self.load is None: self.file: str = self.DEFAULT_FILE
        self.load_session()

    def load_session(self, sid: str = None) -> any:
        if self.file is None and self.load is None and sid is None:
            raise Exception("Neither Load Function nor File Name was passed to load SessionId.")

        if sid:
            try: super().__init__(sid, self.proxy)
            except SessionError: return self.new_session()

        elif self.load:
            session_data: str = self.load().strip()

            if session_data == "": return self.new_session()
            else:
                try: super().__init__(session_data, self.proxy)
                except SessionError: return self.new_session()

        elif self.file:
            if not os.path.exists(self.file): return self.new_session()

            with open(self.file, "r") as reading:
                if (session_data := reading.read().strip()) == "": return self.new_session()

                else:
                    # noinspection PyBroadException
                    try:
                        if json.loads(session_data)["identifier"] != self.identifier: raise Exception()
                        super().__init__(session_data, self.proxy)
                    except Exception: return self.new_session()

    def new_session(self) -> None:
        session_data: str = new_session_id(
            user_identifier=self.identifier,
            password=self.password,
            proxy=self.proxy,
            totp_token=self.totp_token
        )
        
        if self.save: self.save(session_data)

        if self.file:
            with open(self.file, "w") as writing: writing.write(session_data)

        self.load_session(session_data)



import json
from json import JSONDecodeError
from requests import Session, Response
import os
from uuid import uuid4
from .PasswordEncryption import PasswordEncryption
from .lib.Exceptions import AuthenticationError
from requests.models import CaseInsensitiveDict
import pyotp
import ntplib


class Credentials:
    """
    Takes identifier-password and returns Credentials.

    - Should be called with identifier and password.
    - If stored session exists, validates it (not authenticates), and returns Credentials.
    - Otherwise logs in using given identifier and password, stores it in file, and returns Credentials.
    - Raises appropriate exceptions where needed.
    """

    session: Session
    user_agent: str
    totp_token: str

    bearer: str
    user_id: str
    username: str
    phone_id: str
    stored_identifier: str

    device_id: str = "android-d105ec84a512642c"

    def __init__(
        self,
        identifier: str,
        password: str,
        user_agent: str,
        session: Session,
        save_file: str,
        totp_token: str,
        session_data: str,
        logging: bool = False
    ) -> None:

        self.session = session
        self.user_agent = user_agent
        self.totp_token = totp_token

        # Existing Stored Credentials?
        if session_data is not None: stored_dict: dict = json.loads(session_data)
        else: stored_dict: dict = self.fetch_stored_dict(save_file)

        stored_dict_valid: bool = True

        # Validate StoredDict
        for key in ("bearer", "user_id", "username", "phone_id", "identifier", "device_id"):
            if key not in stored_dict.keys(): stored_dict_valid: bool = False

        # Load From StoredDict
        if stored_dict_valid and stored_dict.get("identifier") == identifier:
            if logging: print("Loading from stored file...")
            self.bearer = stored_dict.get("bearer")
            self.user_id = stored_dict.get("user_id")
            self.username= stored_dict.get("username")
            self.phone_id = stored_dict.get("phone_id")
            self.stored_identifier = stored_dict.get("identifier")

        # Create New Session
        else:
            if logging: print("Creating new session...")
            self.login(identifier, password, save_file)

    @staticmethod
    def fetch_stored_dict(save_file: str) -> dict:
        """
        Reads the stored session file content and returns its JSON.
        :param save_file: File Path
        :return: JSON Content or None
        """

        # File Doesn't Exist
        if not os.path.exists(save_file) or not os.path.isfile(save_file): return dict()

        # Read File Content
        with open(save_file, "r") as file:
            content: str = file.read().strip()

        # Is Content A Valid JSON?
        try: return json.loads(content)

        # File Content Not A Valid JSON
        except JSONDecodeError: return dict()


    def login(self, identifier: str, password: str, save_file: str) -> None:
        """
        Takes identifier and password, authenticates, stores new session in file, and returns a bunch of session data.
        :param identifier: Username or Email
        :param password: Password
        :param save_file: File path to store new session in. Empty to skip storing session.
        :return: Tuple of session data
        """

        phone_id: str = str(uuid4())
        guid: str = str(uuid4())
        encrypted_password = PasswordEncryption(self.session).encrypt(password)

        response: Response = self.session.post(
            url="https://i.instagram.com/api/v1/accounts/login/",
            headers={
                "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
                "user-agent": self.user_agent,
                "host": "i.instagram.com",
            },
            data={
                "jazoest": "22506",
                "country_codes": "[{\"country_code\":\"1\",\"source\":[\"default\"]}]",
                "phone_id": phone_id,
                "enc_password": encrypted_password,
                "username": identifier,
                "adid": "00000000-0000-0000-0000-000000000000",
                "guid": guid,
                "device_id": self.device_id,
                "google_tokens": "[]",
                "login_attempt_count": "0"
            }
        )

        try:
            user_info: dict | None = None
            response_headers: CaseInsensitiveDict[str] | None = None

            # Parse Response Body Into A JSON
            response_dict: dict = response.json()

            # Request Failed
            if response_dict.get("status", "fail") != "ok":

                # Other Than 2FA?
                if response_dict.get("error_type", "") != "two_factor_required":
                    raise AuthenticationError(
                        f"Login failed with given credentials.\nError: \"{response_dict.get('error_title', 'unknown')}\"\n"
                        f"Message: \"{response_dict.get('message', 'unknown')}\""
                    )

                # 2FA Required
                user_info, response_headers = self.handle_2fa(
                    information=response_dict.get("two_factor_info", {}),
                    phone_id=phone_id,
                    guid=guid
                )

            # Request Succeeded: Collect Data
            if user_info is None: user_info: dict = response_dict.get("logged_in_user", dict())

            if response_headers is None: self.bearer: str = response.headers.get("ig-set-authorization")
            else: self.bearer: str = response_headers.get("ig-set-authorization")

            self.user_id: str = str(user_info.get("pk"))
            self.phone_id = phone_id
            self.username: str = user_info.get("username")
            self.stored_identifier = identifier

            # Save Session in File
            if save_file != "":
                with open(save_file, "w") as file:
                    file.write(
                        json.dumps({
                            "bearer": self.bearer,
                            "user_id": self.user_id,
                            "username": self.username,
                            "identifier": identifier,
                            "phone_id": phone_id,
                            "device_id": self.device_id
                        })
                    )

        # Response Body Not A Valid JSON
        except JSONDecodeError:
            raise AuthenticationError(
                "Response I got didn't parse into a JSON. Maybe you're being rate "
                "limited. Change WiFi or use reputed proxies."
            )

    def handle_2fa(
        self,
        information: dict,
        phone_id: str,
        guid: str
    ) -> tuple[dict, CaseInsensitiveDict[str]]:

        # Only TOTP Supported: No SMS, No WhatsApp
        if information.get("totp_two_factor_on", False) is False:
            raise AuthenticationError(
                "Only TOTP-based 2FA is supported till now."
            )

        # TODO: Remove when implemented
        raise NotImplementedError("TOTP 2FA not implemented yet!")

        # TOTP Token Supplied? Take Input | Use That
        if self.totp_token is None: code: str = input("Enter TOTP Code: ")
        else: code: str = self.new_totp_code(self.totp_token)

        while True:
            user_info, response_headers = self.bypass_totp(
                code=code,
                two_factor_identifier=information.get("two_factor_identifier"),
                phone_id=phone_id,
                guid=guid
            )

            if user_info is not None and response_headers is not None: break

            if self.totp_token is None:
                code: str = input(
                    "Unable to verify code.\n\n"
                    "Enter TOTP Code: "
                )

            else:
                raise AuthenticationError(
                    "Unable to verify TOTP 2FA Code. Is the totp_token correct?"
                )

        return user_info, response_headers

    @staticmethod
    def new_totp_code(token: str) -> str:
        """
        Generates a new TOTP Code for the current time using secret token.

        :param token: TOTP Token generated by Instagram stored secretly in the Authenticator App.
        :return: Generated Code
        """

        current_time: int = int(ntplib.NTPClient().request("time.google.com", version=3).tx_time)

        return str(
            pyotp.TOTP(token).at(current_time)
        )

    def bypass_totp(self, code: str, two_factor_identifier: str, phone_id: str, guid: str) -> tuple[dict, CaseInsensitiveDict[str]]:

        response: Response = self.session.post(
            url="https://i.instagram.com/api/v1/accounts/two_factor_login/",
            headers={
                "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
                "accept-encoding": "gzip",
                "connection": "Keep-Alive"
            },
            data={
                "signed_body": "SIGNATURE." + json.dumps(
                    {
                        "verification_code": code,
                        "phone_id": phone_id,
                        "two_factor_identifier": two_factor_identifier,
                        "trust_this_device": "1",
                        "guid": guid,
                        "device_id": self.device_id,
                        "waterfall_id": str(uuid4()),
                        "verification_method": "3"  # '3' For TOTP-Based 2FA
                    }
                )
            }
        )

        # TODO
        """
        Return logged_in_user_info and response headers.
        Error: CSRF Token missing or incorrect.
        """

        raise Exception(response.json())



from ensta.structures.Profile import Profile
from .BiographyLinkParser import parse_biography_link


def parse_profile(information: dict) -> Profile:

    user: dict = information.get("user", {})

    profile = Profile(
        full_name=user.get("full_name"),
        username=user.get("username"),
        is_private=user.get("is_private"),
        user_id=user.get("pk"),
        is_profile_picture_expansion_enabled=user.get("is_profile_picture_expansion_enabled"),
        is_opal_enabled=user.get("is_opal_enabled"),
        is_verified=user.get("is_verified"),
        profile_pic_url=user.get("profile_pic_url"),
        biography=user.get("biography"),
        bio_links=tuple(
            parse_biography_link(link) for link in user.get("bio_links", [])
        ),
        account_type=user.get("account_type"),
        is_business=user.get("is_business"),
        media_count=user.get("media_count"),
        following_count=user.get("following_count"),
        follower_count=user.get("follower_count")
    )

    return profile



from requests import Session, Response
from .Credentials import Credentials
from .lib.Exceptions import AuthenticationError, FileTypeError, NetworkError
from uuid import uuid4
from json import  JSONDecodeError
from pathlib import Path
import time
import random
import string
import json
from .parser.ProfileParser import parse_profile
from .structures import Profile
from .Direct import Direct


class Mobile:

    session: Session
    credentials: Credentials

    bearer: str
    user_id: str
    username: str
    phone_id: str
    device_id: str
    identifier: str

    credentials_refresh_max_cycle: int = 3

    authorization_url: str = "https://i.instagram.com/api/v1/wwwgraphql/ig/query/"

    user_agent: str = "Instagram 317.0.0.24.109 Android (31/12; 640dpi; 1440x3056; " \
                      "OnePlus; GM1911; OnePlus7Pro; qcom; en_US; 562739837)"

    def __init__(
        self,
        identifier: str = None,
        password: str = None,
        proxy: dict[str, str] = None,
        save_file: str = "ensta-mobile-session.txt",
        skip_authorization: bool = False,
        logging: bool = False,
        totp_token : str = None,
        session_data: str = None
    ) -> None:

        self.session = Session()

        if proxy: self.session.proxies.update(proxy)

        self.refresh_credentials(
            cycle=1,
            identifier=identifier,
            password=password,
            save_file=save_file,
            logging=logging,
            totp_token=totp_token,
            session_data=session_data,
            skip_authorization=skip_authorization
        )

    def refresh_credentials(
        self,
        cycle: int,
        identifier: str,
        password: str,
        save_file: str,
        logging: bool,
        totp_token: str,
        session_data: str,
        skip_authorization: bool
    ) -> None:

        if logging: print(f"Login Attempt: {cycle} (Max: {self.credentials_refresh_max_cycle})")

        if cycle >= self.credentials_refresh_max_cycle:
            raise AuthenticationError(
                "Reached max number of login attempts while authorizing credentials. "
                "Make sure your network is stable. If required, use reputed proxies."
            )

        self.credentials = Credentials(
            identifier=identifier,
            password=password,
            session=self.session,
            user_agent=self.user_agent,
            save_file=save_file,
            totp_token=totp_token,
            session_data=session_data,
            logging=logging
        )

        self.bearer = self.credentials.bearer
        self.user_id = self.credentials.user_id
        self.username = self.credentials.username
        self.phone_id = self.credentials.phone_id
        self.identifier = self.credentials.stored_identifier
        self.device_id = self.credentials.device_id

        self.session.headers.update({
            "authorization": self.credentials.bearer,
            "host": "i.instagram.com",
            "ig-intended-user-id": self.user_id,
            "ig-u-ds-user-id": self.user_id,
            "user-agent": self.user_agent,
            "x-ig-device-id": self.phone_id,
            "x-ig-device-locale": "en_US",
            "x-ig-family-device-id": self.phone_id,
            "x-fb-connection-type": "WIFI",
            "x-fb-http-engine": "Tigon-TCP-Fallback",
            "x-ig-android-id": self.device_id,
        })

        # Authorization: Is the current session even valid?
        if not skip_authorization:
            if not self.authorize():
                self.refresh_credentials(
                    cycle=cycle + 1,
                    identifier=identifier,
                    password=password,
                    save_file=save_file,
                    logging=logging,
                    totp_token=totp_token,
                    session_data=session_data,
                    skip_authorization=skip_authorization
                )

    def authorize(self) -> bool:
        return self.session.post(self.authorization_url).status_code == 400

    def get_upload_id(self, media_path: str, arg_upload_id: str | None = None) -> str:  # Web API
        """
        Uploads the image to Instagram's server using Web API and returns its UploadID.
        :param media_path: Path to the image file (only jpg & jpeg)
        :param arg_upload_id: Custom upload_id (for advanced users)
        :return: UploadID of picture
        """

        media_path: Path = Path(media_path)

        if media_path.suffix not in (".jpg", ".jpeg"):
            raise FileTypeError(
                "Only jpg and jpeg image types are allowed to upload."
            )

        upload_id = arg_upload_id if arg_upload_id is not None else str(int(time.time()) * 1000)
        upload_name = f"{upload_id}_0_{random.randint(1000000000, 9999999999)}"

        rupload_params = {
            "retry_context": "{\"num_step_auto_retry\": 0, \"num_reupload\": 0, \"num_step_manual_retry\": 0}",
            "media_type": "1",
            "xsharing_user_ids": "[]",
            "upload_id": upload_id,
            "image_compression": json.dumps({"lib_name": "moz", "lib_version": "3.1.m", "quality": 80})
        }

        with open(media_path, "rb") as file:
            photo_data = file.read()
            photo_length = str(len(photo_data))

        request_headers = {
            "accept-encoding": "gzip",
            "x-instagram-rupload-params": json.dumps(rupload_params),
            "x_fb_photo_waterfall_id": str(uuid4()),
            "x-entity-type": "image/jpeg",
            "offset": "0",
            "x-entity-name": upload_name,
            "x-entity-length": photo_length,
            "content-type": "application/octet-stream",
            "content-length": photo_length
        }

        http_response = self.session.post(
            url=f"https://i.instagram.com/rupload_igphoto/{upload_name}",
            data=photo_data,
            headers=request_headers
        )

        try:
            response_dict: dict = http_response.json()

            if response_dict.get("status", "") != "ok": raise NetworkError("Response json key 'status' not ok.")
            if response_dict.get("upload_id") is None:
                raise NetworkError(
                    "Key 'upload_id' in response json doesn't exist. Make sure there's "
                    "nothing wrong with the image you're uploading."
                )

            return str(response_dict.get("upload_id"))

        except JSONDecodeError: raise NetworkError("Response not a valid json.")

    def change_profile_picture(self, image_path: str) -> bool:
        """
        Changes your profile picture to a new one.
        :param image_path: Path of image to be used
        :return: Boolean (Successfully changed or not)
        """

        upload_id: str = self.get_upload_id(image_path)

        headers: dict = {
            "accept-encoding": "gzip",
            "accept-language": "en-US",
            "connection": "Keep-Alive",
            "content-length": "1",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "priority": "u=3",
            "x-bloks-is-layout-rtl": "false",
            "x-bloks-is-prism-enabled": "false",
            "x-bloks-version-id": "979a6b0480455edae83004a50ceae0a15cbe5e943d3786ec785eb85b693c5300",
            "x-ig-app-id": "567067343352427"
        }

        try:
            response: Response = self.session.post(
                "https://i.instagram.com/api/v1/accounts/change_profile_picture/",
                data={
                    "_uuid": str(uuid4()),
                    "use_fbuploader": False,
                    "remove_birthday_selfie": False,
                    "upload_id": upload_id
                },
                headers=headers
            )

            return response.json().get("status", "") == "ok"

        except JSONDecodeError: return False

    def profile(
        self,
        identifier: str
    ) -> Profile:
        """
        Fetches profile information of specified Username or UserID, and returns a Profile Object.
        :param identifier: Username or UserID of target
        :return: Profile Object
        """

        response: Response = self.session.post(
            url=f"https://i.instagram.com/api/v1/users/{identifier}/"
                f"{'username' if self.is_username(identifier) else ''}info_stream/",
            headers={
                "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
                "accept-encoding": "gzip",
                "accept-language": "en-US",
                "connection": "Keep-Alive"
            },
            data={
                "is_prefetch": False,
                "entry_point": "profile",
                "from_module": "search_typeahead",
                "_uuid": str(uuid4())
            }
        )

        try:
            # Response actually returns two JSON Objects with profile information,
            # but we'll only use the 1st one for now.

            information: dict = tuple(json.loads(data) for data in response.text.strip().split("\n"))[0]

            if information.get("status", "") != "ok":
                raise NetworkError(
                    "Cannot fetch profile. Is the username correct? "
                    "Maybe try logging in with a different account as "
                    "this one may has been flagged."
                )

            return parse_profile(information)

        except JSONDecodeError:
            raise NetworkError(
                "Can't fetch profile info. Make sure username or user-id you supplied is correct. "
                "If this issue persists, try using a different account as probably the current one has "
                "been flagged."
            )

    @staticmethod
    def is_username(identifier: str) -> bool:
        """
        Returns True if the given identifier is a Username, and False if it's a UserID.
        :param identifier: Username or UserID
        :return: Boolean
        """

        return False in tuple(x in string.digits for x in identifier)

    def direct(self) -> Direct:
        """
        Lets you use the direct messaging feature of Instagram
        :return: Direct() class with all the supported methods
        """

        return Direct(
            session=self.session,
            device_id=self.device_id
        )

    def follow(self, user_id: str) -> bool:
        """
        Follows a user by user_id
        :param self: None
        :param user_id: UserID of target user
        :return: Boolean (Followed or not)
        """

        response: Response = self.session.post(
            url=f"https://i.instagram.com/api/v1/friendships/create/{user_id}/",
            data="SIGNATURE." + json.dumps(
                {
                    "user_id": user_id,
                    "radio_type": "wifi-none",
                    "device_id": self.device_id,
                    "_uuid": str(uuid4()),
                    "container_module": "profile"
                }
            )
        )

        try: return response.json().get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError(
                "Unable to follow. Is the user_id correct? Try using another account, switch "
                "to a different network, or use reputed proxies."
            )

    def unfollow(self, user_id: str) -> bool:
        """
        Unfollows a user by user_id
        :param self: None
        :param user_id: UserID of target user
        :return: Boolean (Unfollowed or not)
        """

        response: Response = self.session.post(
            url=f"https://i.instagram.com/api/v1/friendships/destroy/{user_id}/",
            data="SIGNATURE." + json.dumps(
                {
                    "user_id": user_id,
                    "radio_type": "wifi-none",
                    "device_id": self.device_id,
                    "_uuid": str(uuid4()),
                    "container_module": "profile"
                }
            )
        )

        try: return response.json().get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError(
                "Unable to follow. Is the user_id correct? Try using another account, switch "
                "to a different network, or use reputed proxies."
            )

    def block(self, user_id: str) -> bool:
        """
        Blocks a user by user_id
        :param self: None
        :param user_id: UserID of target user
        :return: Boolean (Blocked or not)
        """

        response: Response = self.session.post(
            url=f"https://i.instagram.com/api/v1/friendships/block/{user_id}/",
            data="SIGNATURE." + json.dumps(
                {
                    "surface": "profile",
                    "is_auto_block_enabled": True,
                    "user_id": user_id,
                    "_uid": self.user_id,
                    "_uuid": str(uuid4()),
                    "container_module": "profile"
                }
            )
        )

        try: return response.json().get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError(
                "Unable to block. Is the user_id correct? Username should not be used. Try using another account, switch "
                "to a different network, or use reputed proxies."
            )

    def unblock(self, user_id: str) -> bool:
        """
        Unblocks a user by user_id
        :param self: None
        :param user_id: UserID of target user
        :return: Boolean (Unblocked or not)
        """

        response: Response = self.session.post(
            url=f"https://i.instagram.com/api/v1/friendships/unblock/{user_id}/",
            data="SIGNATURE." + json.dumps(
                {
                    "user_id": user_id,
                    "_uid": self.user_id,
                    "_uuid": str(uuid4()),
                    "container_module": "search_typeahead"
                }
            )
        )

        try: return response.json().get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError(
                "Unable to unblock. Is the user_id correct? Username should not be used. Try using another account, switch "
                "to a different network, or use reputed proxies."
            )
